

class AS::Texture
  # The width and height of the texture.
  attr_reader :size

  # OpenGL texture object id.
  attr_reader :tex_id

  # Texture image width.
  attr_reader :w

  # Texture image height.
  attr_reader :h
  
  # Possible _filter_ parameters.
  Filter = {
    :smooth  => [GL::LINEAR, GL::LINEAR],
    :blocky => [GL::NEAREST, GL::NEAREST],
    :mip_smooth => [GL::LINEAR_MIPMAP_LINEAR, GL::LINEAR],
    :mip_blocky => [GL::NEAREST_MIPMAP_LINEAR, GL::LINEAR],
  }
  
  # Possible _inform_ parameters
  Format = {
    :alpha => GL::ALPHA,
    :bw    => GL::LUMINANCE,
    :rgb   => GL::RGB,
    :rgba  => GL::RGBA,
  }
  
  def pot?(num)
    0.upto(16){|n| return true if num == 2**n}
    return false
  end
  private :pot?

  TextureDimError = Class.new(Exception)

  def validate(image)
    unless pot?(image.w) and pot?(image.h)
      raise(TextureDimError, "Width or height not a power of 2")
    end
  end
  private :validate

  def initialize(image, filter = :smooth, inform = :rgba, repeat = false)
    validate(image)
    @pixels = image.data
    @w = image.w
    @h = image.h
    @pix_format = image.format
    
    @filter = case filter
    when Symbol: Filter[filter]
    when Array: filter
    end
	
    raise(ArgumentError, "#{filter} is not a valid texture filter") unless @filter    

    @repeat = case repeat
    when true: GL::REPEAT
    when false: GL::CLAMP
    when Fixnum: repeat
    else raise(ArgumentError, "repeat must be a GL constant or boolean.")
    end
    
    @inform = case inform
    when Symbol: Format[inform]
    when Fixnum: inform
    end
	
    raise(ArgumentError, "#{inform} is not a valid texture format") unless @inform

    @tex_id = 0
	
	return load
  end
  
  def self.finalizer_block(id)
		lambda{GL.DeleteTextures([id])}
  end

  def gen_tex_id
    @tex_id = GL.GenTextures(1).first
    ObjectSpace.define_finalizer(self, AS::Texture.finalizer_block(tex_id))
  end
  private :gen_tex_id
  
  def load
    gen_tex_id
    bind
	
    GL.TexParameter(tex_target, GL::TEXTURE_WRAP_S, @repeat)
    GL.TexParameter(tex_target, GL::TEXTURE_WRAP_T, @repeat)
	minf, magf = *@filter
    GL.TexParameter(tex_target, GL::TEXTURE_MIN_FILTER, minf)
    GL.TexParameter(tex_target, GL::TEXTURE_MAG_FILTER, magf)
    GL.TexEnv(GL::TEXTURE_ENV, GL::TEXTURE_ENV_MODE, tex_env_mode)
    
    upload
    	    
    return self
  end
  private :load
  
  def upload
    GL.TexImage2D(tex_target, 0, @inform, @w, @h, 0, @pix_format, GL::UNSIGNED_BYTE, @pixels)
  end
  private :upload

  # Set as the current texture unit.
  def bind
    GL::BindTexture(tex_target, @tex_id)
  end

  # Draw from the source rectangle in texture coordinates into the
  # destination rectangle in world coordinates.
  def copy_rect(sx, sy, sw, sh, dx, dy, dw, dh)
    sright = sx + sw
    stop   = sy + sh

    dright = dx + dw
    dtop   = dy + dh

    GL.Begin(GL::QUADS)
		GL.TexCoord2f(    sx,   sy); GL.Vertex2f(    dx,   dy)
		GL.TexCoord2f(sright,   sy); GL.Vertex2f(dright,   dy)
		GL.TexCoord2f(sright, stop); GL.Vertex2f(dright, dtop)
		GL.TexCoord2f(    sx, stop); GL.Vertex2f(    dx, dtop)
    GL.End
  end
  
  # Draw from the source rectangle in texture coordinates into the
  # destination rectangle in world coordinates.
  def self.copy_rect(sx, sy, sw, sh, dx, dy, dw, dh)
    sright = sx + sw
    stop   = sy + sh

    dright = dx + dw
    dtop   = dy + dh

    GL.Begin(GL::QUADS)
		GL.TexCoord2f(    sx,   sy); GL.Vertex2f(    dx,   dy)
		GL.TexCoord2f(sright,   sy); GL.Vertex2f(dright,   dy)
		GL.TexCoord2f(sright, stop); GL.Vertex2f(dright, dtop)
		GL.TexCoord2f(    sx, stop); GL.Vertex2f(    dx, dtop)
    GL.End
  end
  
  # Returns the default OpenGL texture environment mode. Override in
  # subclasses to use different modes.
  def tex_env_mode
    GL::MODULATE
  end
  
  # Returns the default texture target. Overriden in AS::RectTexture
  # to provide support for rectangle textures.
  def tex_target
    GL::TEXTURE_2D
  end
  
  def self.open(filename)
    new(AS::Image.load_img(filename))
  end
end

# Rectangle texture class. Rectangle textures do not have the same
# size limitations as POT textures, but don't support repeating.
class AS::RectTexture < AS::Texture
  def initialize(backing, filter = :smooth, inform = :rgba)
    super(backing, filter, inform, false)
  end

  def validate(image)
    return true
  end
  private :validate
  
  def tex_target
    return GL::TEXTURE_RECTANGLE_EXT
  end
end

# MIP textures are a special kind of POT texture that use mipmaps and
# trilinear filtering to provide smoother scaling at small sizes. Uses
# 25-33% more memory than a POT texture of the same size.
class AS::MIPTexture < AS::Texture
  def initialize(backing, filter = :mip_smooth, inform = :rgba, repeat = false)
    super
  end

  def upload
    GLU.Build2DMipmaps(tex_target, @inform, @w, @h, @pix_format, GL::UNSIGNED_BYTE, @pixels)
  end
end
